#include <string>
#include <fstream>
#include <new>
#include <iostream>

#include "MyTmpltLib.h"

using namespace std;

#include "Course.cpp"
#include "Student.cpp"
#include "StudentInCourse.cpp"


struct ProgXptn
{
   enum { allocFailed, cannotOpen, cannotRead };
   char functionName[100];
   int  errorCode;
   char errStrVal[100];
   ProgXptn( const char* fn, int ec, const char* sv="" )
   {  strncpy( functionName, fn, 99 );  functionName[99] = '\0';
      errorCode = ec;
      strncpy( errStrVal, sv, 99 );  errStrVal[99] = '\0';  }
}; // ProgXptn

void loadCourses( string flNm, PCourse& crsArr, unsigned int& nOfCourses );
void readAStudent( istream& tin,
                   PStudent& stArr, unsigned int& nOfStudents,
                   unsigned int& resStudents, int sIncr );
void ignoreStudentData( istream& tin );
void readCourseCodes( istream& tin, Student& aStudent,
                      PCourse crsArr, int nOfCourses,
                      PStudentInCourse& sc, unsigned int& nOfStdInCrs,
                      unsigned int& resStdInCrs, int scIncr, ostream& log );

int main()
{
   const int maxWH( 30 );
   PCourse   crsArr;  unsigned int nOfCourses;

   PStudent  stArr;   unsigned int nOfStudents( 0 );
   const int sIncr( 10 );
   unsigned int resStudents;

   PStudentInCourse sc; unsigned int nOfStdInCrs( 0 );
   const int scIncr( 30 );
   unsigned int resStdInCrs;

   ofstream log;
   try
   {
      loadCourses( "gCourses.dta", crsArr, nOfCourses );

      try { stArr = new Student[sIncr]; }
      catch( bad_alloc )
      {  throw ProgXptn( "main", ProgXptn::allocFailed );  }
      resStudents = sIncr;
      try { sc = new StudentInCourse[scIncr]; }
      catch( bad_alloc )
      {  throw ProgXptn( "main", ProgXptn::allocFailed );  }
      resStdInCrs = scIncr;

      log.open( "log.txt" );
      if ( log.fail() )
         throw ProgXptn( "main", ProgXptn::cannotOpen, "log.txt" );
      ifstream tin( "enrllmnt.txt" );
      if ( tin.fail() )
         throw ProgXptn( "main", ProgXptn::cannotOpen, "enrllmnt.txt" );
      while ( !tin.eof() )
      {
         readAStudent( tin, stArr, nOfStudents, resStudents, sIncr );
         if ( !tin.eof() )
         {
            int pos( 0 );
            while ( stArr[pos].getIdNum() !=
                    stArr[nOfStudents-1].getIdNum() ) ++pos;
            if ( pos < nOfStudents-1 )  //  
            {//       
               --nOfStudents;
               ignoreStudentData( tin );
               log << " multiple entry for student with id num "
                   << stArr[pos].getIdNum() << endl;
            }
            else  //  
            {
               readCourseCodes( tin, stArr[nOfStudents-1], crsArr, nOfCourses,
                                sc, nOfStdInCrs, resStdInCrs, scIncr, log );
            }
         }
         if ( !tin.eof() )
         {
            string str1;
            getline( tin, str1, '\n' ); // blank line
         }
      } // while
      tin.close();
      for ( int k(0); k < nOfStudents; ++k )
      {
         if ( stArr[k].getWH() > maxWH )
            log << "student with id num " << stArr[k].getIdNum()
                << ": " << stArr[k].getWH() << " hours/week" << endl;
      } // for
      save1DTable( "Students.dta", stArr, nOfStudents );
      save1DTable( "enrllmnt.dta", sc, nOfStdInCrs );
      save1DTable( "Courses.dta", crsArr, nOfCourses );
      delete[] sc;
      delete[] stArr;
      delete[] crsArr;
   } // try
   catch( ProgXptn& x )
   {
      switch ( x.errorCode )
      {
         case ProgXptn::allocFailed:
            cout << "no dynamic memory in " << x.functionName << endl;
            break;
         case ProgXptn::cannotOpen:
            if ( strcmp(x.errStrVal, "log.txt") == 0 )
               cout << "cannot open file enrllmnt.txt" << endl;
            else
               log << "cannot open file " << x.errStrVal << " in "
                   << x.functionName << endl;
            break;
         default:
            cout << "unexpected ProgXptn from " << x.functionName << endl;
      } // switch
   } // catch( ProgXptn
   catch( MyTmpltLibXptn& x )
   {
      switch ( x.errCode )
      {
         case MyTmpltLibXptn::noArray:
            cout << x.funcName << " called with NULL array pointer"
                 << endl;
            break;
         case MyTmpltLibXptn::paramError:
            cout << x.funcName << ": param error ni(" << x.errIntVal1
                 << ") and nf(" << x.errIntVal2 << ")" << endl;
            break;
         case MyTmpltLibXptn::allocFailed:
            cout << "no dynamic memory in " << x.funcName << endl;
            break;
         case MyTmpltLibXptn::cannotCreate:
            cout << "cannot create file " << x.errStrVal << " in "
                 << x.funcName << endl;
            break;
         case MyTmpltLibXptn::cannotWrite:
            cout << "cannot write to file " << x.errStrVal << " in "
                 << x.funcName << endl;
            break;
         default:
            cout << "unexpected MyTmpltLibXptn from "
                 << x.funcName << endl;
      } // switch
   } // catch( MyTmpltLibXptn
   catch( CourseXptn& x )
   {
      switch ( x.errorCode )
      {
         case CourseXptn::fileNotOpen:
            cout << "file " << x.errStrVal << " not open in "
                 << x.methodName << endl;
            break;
         case CourseXptn::cannotWrite:
            cout << "cannot write to file " << x.errStrVal << " in "
                 << x.methodName << endl;
            break;
         case CourseXptn::cannotRead:
            cout << "cannot read from file " << x.errStrVal << " in "
                 << x.methodName << endl;
            break;
         default:
            cout << "unexpected CourseXptn from " << x.methodName << endl;
      } // switch
   } // catch( CourseXptn
   catch( StudentXptn& x )
   {
      switch ( x.errorCode )
      {
         case StudentXptn::negIdNum:
            cout << "negative Student id num (" << x.errIntVal << ") in "
                 << x.methodName << endl;
            break;
         case StudentXptn::nonPosIncr:
            cout << "non positive increment (" << x.errIntVal << ") in "
                 << x.methodName << endl;
            break;
         case StudentXptn::fileNotOpen:
            cout << "file " << x.errStrVal << " not open in "
                 << x.methodName << endl;
            break;
         case StudentXptn::cannotWrite:
            cout << "cannot write to file " << x.errStrVal << " in "
                 << x.methodName << endl;
            break;
         case StudentXptn::cannotRead:
            cout << "cannot read from file " << x.errStrVal << " in "
                 << x.methodName << endl;
            break;
         default:
            cout << "unexpected StudentXptn from " << x.methodName << endl;
      } // switch
   } // catch( StudentXptn
   catch( StudentInCourseXptn& x )
   {
      switch ( x.errorCode )
      {
         case StudentInCourseXptn::unknownCCode:
            cout << "unknown Course Code (" << x.errStrVal << ") in "
                 << x.methodName << endl;
            break;
         case StudentInCourseXptn::fileNotOpen:
            cout << "file " << x.errStrVal << " not open in "
                 << x.methodName << endl;
            break;
         case StudentInCourseXptn::cannotWrite:
            cout << "cannot write to file " << x.errStrVal << " in "
                 << x.methodName << endl;
            break;
         default:
            cout << "unexpected StudentInCourseXptn from "
                 << x.methodName << endl;
      } // switch
   } // catch( StudentInCourseXptn
   catch( ... )
   {
      cout << "unexpected exceptioon" << endl;
   }
   log.close();
system("pause");
} // main

void loadSylCourse( ifstream& bin, Course& oneCourse )
{
   bool         bBuf;
   unsigned int iBuf;
   char         cBuf[Course::cTitleSz];

   bin.read( cBuf, Course::cCodeSz );         //  
   if ( !bin.eof() )
   {
                                        oneCourse.setCode( cBuf );
      bin.read( cBuf, Course::cTitleSz );      //  
                                       oneCourse.setTitle( cBuf );
      bin.read( reinterpret_cast<char*>(&iBuf),
                sizeof(unsigned int) );          //  
                                        oneCourse.setFSem( iBuf );
      bin.read( reinterpret_cast<char*>(&bBuf),
                sizeof(bool) );          //   
                                     oneCourse.setCompuls( bBuf );
      bin.read( cBuf, sizeof(char) );                    // 
                                   oneCourse.setSector( cBuf[0] );
      bin.read( cBuf, Course::cCategSz );             // 
                                       oneCourse.setCateg( cBuf );
      bin.read( reinterpret_cast<char*>(&iBuf),
                sizeof(unsigned int) );       //   
                                          oneCourse.setWH( iBuf );
      bin.read( reinterpret_cast<char*>(&iBuf),
                sizeof(unsigned int) );      //  
                                       oneCourse.setUnits( iBuf );
      bin.read( cBuf, Course::cCodeSz );         // 
                                      oneCourse.setPrereq( cBuf );
      if ( bin.fail() )
         throw ProgXptn( "loadSylCourse", ProgXptn::cannotRead );
   }
} // loadSylCourse

void loadCourses( string flNm, PCourse& crsArr, unsigned int& nOfCourses )
{
//  () 
   const int cIncr( 10 );
   unsigned int resCourses;
   crsArr = 0;   nOfCourses = 0;
   try {  crsArr = new Course[cIncr];  }
   catch( bad_alloc )
   {  throw ProgXptn( "loadCourses", ProgXptn::allocFailed );  }
   resCourses = cIncr;
//   
   ifstream bin( flNm.c_str(), ios_base::binary );
   if ( bin.fail() )
   {
      delete[] crsArr;   //   
      throw ProgXptn( "loadCourses", ProgXptn::cannotOpen, flNm.c_str() );
   }
//  
   loadSylCourse( bin, crsArr[nOfCourses] );
   while ( !bin.eof() )
   {
      ++nOfCourses;
      crsArr[nOfCourses-1].clearStudents();
      if ( nOfCourses+1 == resCourses )
      {
         renew( crsArr, nOfCourses, resCourses+cIncr );
         resCourses += cIncr;
      } // if
      loadSylCourse( bin, crsArr[nOfCourses] );
   } // while
   bin.close();
} // loadCourses

void readAStudent( istream& tin,
                   PStudent& stArr, unsigned int& nOfStudents,
                   unsigned int& resStudents, int sIncr )
{
   if ( nOfStudents+1 == resStudents )
   {
      renew( stArr, nOfStudents, resStudents+sIncr );
      resStudents += sIncr;
   }
   stArr[nOfStudents].readFromText( tin );
   if ( !tin.eof() )
   {
      ++nOfStudents;
      stArr[nOfStudents-1].clearCourses();
//      stArr[nOfStudents-1].clearWH();
   }
} // readAStudent

void ignoreStudentData( istream& tin )
{
   string str1;
   getline( tin, str1, '\n' );
   int noc( atoi(str1.c_str()) );
   for ( int k(0); k < noc; ++k )  getline( tin, str1, '\n' );
} // ignoreStudentData

void readCourseCodes( istream& tin, Student& aStudent,
                      PCourse crsArr, int nOfCourses,
                      PStudentInCourse& sc, unsigned int& nOfStdInCrs,
                      unsigned int& resStdInCrs, int scIncr, ostream& log )
{
   if ( !tin.eof() )
   {
      string str1;
      getline( tin, str1, '\n' );
      if ( !tin.eof() )
      {
         int noc( atoi(str1.c_str()) );
         for ( int k(0); k < noc; ++k )
         {
            if ( nOfStdInCrs+1 == resStdInCrs )
            {
               renew( sc, nOfStdInCrs, resStdInCrs+scIncr );
               resStdInCrs += scIncr;
            }
            getline( tin, str1, '\n' );
            if ( !tin.eof() )
            {
               try
               {
                  sc[nOfStdInCrs].setIdNumCCode( aStudent, str1.c_str(),
                                                 crsArr, nOfCourses );
                  ++nOfStdInCrs;
               }
               catch( StudentInCourseXptn& x )
               {
                  log << " student with id num " << aStudent.getIdNum()
                      << " asking course " << str1 << endl;
               }
            } // if ( !tin.eof()...
         } // for
      }
   }
} // readCourseCodes



